/*----------------------------------------------------------------------------
 * lilprox - Proxies (sort of) HTTP requests and allows modifications via perl
 * Copyright (C) 2005 Arne Bochem <lilprox@arneb.de>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *----------------------------------------------------------------------------
 * Version: 1.2
 *----------------------------------------------------------------------------
 * Compile with:
 *   make
 *----------------------------------------------------------------------------
 * Credits:
 *   Fabian Pietsch (status error handling and stuff)
 *----------------------------------------------------------------------------*/

/* Evil the code is. */
#include <stdio.h>
#include <strings.h>
#include <stdlib.h>
#include <string.h>
#include <signal.h>
#include <sys/socket.h>
#include <sys/wait.h>
#include <sys/types.h>
#include <sys/time.h>
#include <arpa/inet.h>
#include <unistd.h>
#include <netdb.h>

/* Glue */
#include <EXTERN.h>
#include <perl.h>

enum status
{
	LS_SUCCESS = 0,
	LS_BAD_REQUEST,
	LS_NO_DNS,
	LS_NO_CONNECT
};

#define CON_TIMEOUT 60

char *bad_request = "HTTP/1.0 400 Bad Request\nContent-Type: text/html\n\nInvalid URL scheme.\n";
char *no_dns = "HTTP/1.0 503 Service Unavailable\nContent-Type: text/html\n\nCannot resolve host.\n";
char *no_connect = "HTTP/1.0 503 Service Unavailable\nContent-Type: text/html\n\nCannot connect to remote host.\n";
char *timeout = "HTTP/1.0 503 Service Unavailable\nContent-Type: text/html\n\nConnection timed out.\n";
char *ipe = "HTTP/1.0 503 Service Unavailable\nContent-Type: text/html\n\nInternal proxy error.\n";

static PerlInterpreter *my_perl = NULL;
unsigned int child_num = 0;
char *proxy_host = NULL;

void err_client(int client, enum status err)
{
	char *msg = ipe;

	switch (err)
	{
		case LS_SUCCESS:
			return;
		case LS_BAD_REQUEST:
			msg = bad_request;
			break;
		case LS_NO_DNS:
			msg = no_dns;
			break;
		case LS_NO_CONNECT:
			msg = no_connect;
			break;
	}

	write(client, msg, strlen(msg));
	close(client);
	exit(1);
}

void perl_filter (char *sub, char **data, STRLEN *length)
{
	char *temp;
	STRLEN n_a;

	if (my_perl != NULL)
	{
		dSP;
		ENTER;
		SAVETMPS;
		PUSHMARK(SP);
		XPUSHs(sv_2mortal(newSVpvn(*data, *length)));
		PUTBACK;
		if (call_pv(sub, G_SCALAR) != 1)
		{
			fprintf(stderr, "Perl error!");
			exit(1);
		}
		SPAGAIN;
		temp = POPpx;
		*length = n_a;
		free(*data);
		*data = malloc(*length);
		memcpy(*data, temp, *length);
		PUTBACK;
		FREETMPS;
		LEAVE;
	}
}

/* Taken from GNU libc manual. */
int input_timeout (int filedes)
{
	fd_set set;
	struct timeval timeout;

	/* Initialize the file descriptor set. */
	FD_ZERO(&set);
	FD_SET(filedes, &set);

	/* Initialize the timeout data structure. */
	timeout.tv_sec = CON_TIMEOUT;
	timeout.tv_usec = 0;

	/* select returns 0 if timeout, 1 if input available, -1 if error. */
	return TEMP_FAILURE_RETRY(select (FD_SETSIZE, &set, NULL, NULL, &timeout));
}

char *get_content (int client, int content_length)
{
	int to_read, has_read;
	char *con_ptr, *content;

	if ((content = malloc(content_length)) == NULL)
	{
		perror("malloc()");
		abort();
	}

	con_ptr = content;
	to_read = content_length;
	while ((to_read > 0) && (has_read > 0))
	{
		if (input_timeout(client) != 1)
		{
			write(client, timeout, strlen(timeout));
			close(client);
			exit(1);
		}

		if ((has_read = read(client, con_ptr, to_read)) == -1)
		{
			perror("read()");
			abort();
		}

		con_ptr += has_read;
		to_read -= has_read;
	}

	return(content);
}

int get_content_length (char *header)
{
	char *conlen = strstr(header, "\nContent-Length: ") + 17;

	if (conlen == NULL + 17)
		return(0);

	return(atoi(conlen));
}

enum status get_path (char **path, char *url, char *host)
{
	char *path1;
	char *temp;

	if ((path1 = strchr(host, '/')) == NULL)
		return LS_BAD_REQUEST;

	if ((temp = malloc(strlen(path1) + 1)) == NULL)
	{
		perror("malloc()");
		abort();
	}

	memset(temp, 0, strlen(path1) + 1);
	strncpy(temp, path1, strlen(path1));

	path1[0] = '\0';
	*path = temp;

	return(LS_SUCCESS);
}

enum status get_host_and_path (char **host, char *url)
{
	if (strstr(url, "http://") != url)
		return LS_BAD_REQUEST;

	*host = malloc(strlen(url + 7) + 1);
	strcpy(*host, url + 7);

	return(LS_SUCCESS);
}

enum status get_host (char **host, char **header_ptr)
{
	char *url1, *url2, *url;
	char *header = *header_ptr;
	char *path_header;
	char *path = "";
	int h_len, p_len;
	enum status err;

	if (proxy_host != NULL)
	{
		*host = malloc(strlen(proxy_host) + 1);
		strcpy(*host, proxy_host);
		return LS_SUCCESS;
	}

	/* URL */
	if ((url1 = strchr(header, ' ') + 1) == NULL + 1)
		return LS_BAD_REQUEST;
	if ((url2 = strchr(url1, ' ')) == NULL)
		return LS_BAD_REQUEST;

	if ((url = malloc(url2 - url1 + 1)) == NULL)
	{
		perror("malloc()");
		abort();
	}

	memset(url, 0, url2 - url1 + 1);
	strncpy(url, url1, url2 - url1);

	url1[0] = '\0';

	/* HOST */
	if ((err = get_host_and_path(host, url)))
		return err;

	/* PATH */
	if ((err = get_path(&path, url, *host)))
		return err;

	/* BUILD */
	h_len = strlen(header);
	p_len = strlen(path);

	path_header = malloc(h_len + p_len + strlen(url2) + 1);
	strcpy(path_header, header);
	strcpy(path_header + h_len, path);
	strcpy(path_header + h_len + p_len, url2);

	free(header);
	free(path);
	free(url);

	*header_ptr = path_header;

	return(LS_SUCCESS);
}

char *get_header (int socket)
{
	int h_pos = -1;
	char last = 0;
	char *header = NULL;
	int read_v;

	do
	{
		if ((h_pos != -1) && (header[h_pos] != '\r'))
			last = header[h_pos];
		
		h_pos++;
		if (h_pos % 1024 == 0)
		{
			if ((header = (char *)realloc(header, h_pos + 1024)) == NULL)
			{
				perror("realloc()");
				abort();
			}
			memset(header + h_pos, 0, 1024);
		}

		if (input_timeout(socket) != 1)
		{
			/* Sending error messages to servers is great fun. Not? */
			/* write(socket, timeout, strlen(timeout)); */
			return(header);
		}

		if ((read_v = read(socket, header + h_pos, 1)) == -1)
		{
			perror("read()");
			abort();
		}
	}
	while (((last != '\n') || (header[h_pos] != '\n')) && (read_v > 0));

	if ((header = (char *)realloc(header, h_pos + 2)) == NULL)
	{
		perror("realloc()");
		abort();
	}

	return(header);
}

enum status do_request(int client, char *host, char *request, int req_len)
{
	char *port_sep;
	int port = 80;
	int remote;
	struct sockaddr_in r_add;
	struct hostent *r_server;
	char *header;
	int read_v, h_len;
	char *post;
	int p_len = 0;
	char *content;

	/* PORT */
	port_sep = strchr(host, ':');
	if (port_sep != NULL)
	{
		if (strlen(port_sep) == 1)
			return LS_BAD_REQUEST;

		*port_sep = '\0';

		if (strlen(host) == 0)
			return LS_BAD_REQUEST;

		port = atoi(port_sep + 1);
	}

	/* CONNECTION */
	if ((remote = socket(PF_INET, SOCK_STREAM, IPPROTO_TCP)) == -1)
	{
		perror("socket()");
		exit(1);
	}

	if ((r_server = gethostbyaddr(host, strlen(host), AF_INET)) == NULL)
		if ((r_server = gethostbyname(host)) == NULL)
			return LS_NO_DNS;

	memset(&r_add, 0, sizeof(r_add));
	r_add.sin_family = AF_INET;
	memcpy(&r_add.sin_addr.s_addr, r_server->h_addr, r_server->h_length);
	r_add.sin_port = htons(port);

	if (connect(remote, (struct sockaddr *)&r_add, sizeof(r_add)) == -1)
		return LS_NO_CONNECT;

	/* SEND */
	write(remote, request, req_len);

	/* HEADER */
	header = get_header(remote);
	h_len = strlen(header);
	perl_filter("server_header", &header, &h_len);
	write(client, header, h_len);
	/*fprintf(stderr, "%s", header);*/

	/* CONTENT */
	content = malloc(1024);

	if (input_timeout(remote) != 1)
	{
		write(client, timeout, strlen(timeout));
		close(remote);
		close(client);
		exit(1);
	}

	while ((read_v = read(remote, content, 1024)) > 0)
	{
		perl_filter("server_content", &content, &read_v);
		write(client, content, read_v);
		free(content);
		content = malloc(1024);
		if (input_timeout(remote) != 1)
			break;
	}
	/*fprintf(stderr, "%s", content);*/

	post = malloc(1);
	post[0] = '\0';
	perl_filter("server_post", &post, &p_len);
	write(client, post, p_len);

	if (read_v == -1)
	{
		perror("read()");
		abort();
	}

	/* DONE */
	close(remote);
	close(client);

	free(content);
	free(header);
	free(post);

	return LS_SUCCESS;
}

void handle_client (int client)
{
	char *header;
	int content_length = 0;
	int req_len, h_len;
	char *host, *content, *request;
	enum status err;
	pid_t pid = fork();

	if (pid == -1)
	{
		perror("fork()");
		abort();
	}

	if (pid != 0)
	{
		child_num++;
		close(client);
		return;
	}

	/* HEADER */
	header = get_header(client);
	h_len = strlen(header);
	perl_filter("client_header", &header, &h_len);

	/* printf("%s", header); */

	/* HOST */
	if ((err = get_host(&host, &header)))
		err_client(client, err);

	/* fprintf(stderr, ":%s:", host); */

	/* CONTENT LENGTH */
	content_length = get_content_length(header);
 
	/* fprintf(stderr, "!%i!", content_length); */

	/* CONTENT */
	content = get_content(client, content_length);
	perl_filter("client_content", &content, &content_length);

	/* REQUEST */
	h_len = strlen(header);

	if ((request = malloc(h_len + content_length)) == NULL)
	{
		perror("malloc()");
		abort();
	}

	strncpy(request, header, h_len);
	strncpy(request + h_len, content, content_length);
	req_len = h_len + content_length;
	perl_filter("client_request", &request, &req_len);

	/* printf("!%s!", request); */

	/* NETWORK */
	if ((err = do_request(client, host, request, req_len)))
		err_client(client, err);

	free(header);
	free(request);
	free(content);
	free(host);

	exit(0);
}

void child_signal (int arg)
{
	pid_t child;

	while (child_num > 0)
	{
		if ((child = waitpid((pid_t) -1, NULL, WNOHANG)) == -1)
		{
			perror("waitpid()");
			abort();
		}
		else
			if (child == 0)
				break;
			else
				child_num--;
	}
}

void init_perl ()
{
	char *args[] = {"nullperl", "glue.pl"};
	PERL_SYS_INIT3(0, NULL, NULL);
	my_perl = perl_alloc();
	perl_construct(my_perl);
	PL_exit_flags |= PERL_EXIT_DESTRUCT_END;
	perl_parse(my_perl, NULL, 2, args, NULL);
}

void kill_perl ()
{
	PerlInterpreter *perl = my_perl;
	my_perl = NULL;
	perl_destruct(perl);
	perl_free(perl);
	PERL_SYS_TERM();
}

void hup_signal (int arg)
{
	kill_perl();
	init_perl();
}

int main (int argc, char *argv[])
{
	int server;
	struct sockaddr_in server_add;
	int client;
	struct sockaddr_in client_add;
	int client_len;

	if (argc == 2)
	{
		proxy_host = argv[1];
	}

	if ((server = socket(PF_INET, SOCK_STREAM, IPPROTO_TCP)) == -1)
	{
		perror("socket()");
		abort();
	}

	memset(&server_add, 0, sizeof(server_add));
	server_add.sin_family = AF_INET;
	server_add.sin_addr.s_addr = htonl(INADDR_ANY);
	server_add.sin_port = htons(5432);

	if (signal(SIGCHLD, child_signal) == SIG_ERR)
	{
		perror("signal()");
		abort();
	}

	if (signal(SIGHUP, hup_signal) == SIG_ERR)
	{
		perror("signal()");
		abort();
	}

	if (bind(server, (struct sockaddr *)&server_add, sizeof(server_add)) == -1)
	{
		perror("bind()");
		abort();
	}

	if (listen(server, 10) == -1)
	{
		perror("listen()");
		abort();
	}

	init_perl();
	atexit(kill_perl);

	for (;;)
	{
		client_len = sizeof(client_add);

		if ((client = accept(server, (struct sockaddr *)&client_add,
			(socklen_t *)&client_len)) == -1)
		{
			perror("accept()");
			abort();
		}

		handle_client(client);
	}
}
